#ifndef _ABSTRACTION_HELPERS_HPP_
#define _ABSTRACTION_HELPERS_HPP_

#include <alib/tuple>
#include <alib/array>
#include <alib/typeinfo>

namespace abstraction {

template < class ParamType >
std::shared_ptr < ValueProvider < ParamType > > translateParam ( const std::shared_ptr < OperationAbstraction > & param ) {
	std::shared_ptr < ValueProvider < ParamType > > res = std::dynamic_pointer_cast < ValueProvider < ParamType > > ( param->getProxyAbstraction ( ) );
	if ( ! res )
		throw std::invalid_argument ( "Abstraction does not provide value of type " + ext::to_string < ParamType > ( ) + " but " + param->getReturnType ( ) + "." );
	return res;
}

template < class ParamType >
ParamType retrieveValue ( const std::shared_ptr < OperationAbstraction > & param, bool move ) {
	return translateParam < ParamType > ( param )->getValue ( move );
}

namespace detail {

template < class ... ParamTypes, class F, class Tuple, std::size_t... I >
constexpr decltype ( auto ) apply_impl ( F && f, Tuple && t, std::index_sequence < I ... > ) {
	return std::invoke ( std::forward < F > ( f ), abstraction::retrieveValue < ParamTypes > ( std::get < I > ( std::forward < Tuple > ( t ) ).first, std::get < I > ( std::forward < Tuple > ( t ) ).second ) ... );
}

}  // namespace detail

template < class ... ParamTypes, class F, class Tuple >
constexpr decltype ( auto ) apply ( F && f, Tuple && t ) {
	return detail::apply_impl < ParamTypes ... > ( std::forward < F > ( f ), std::forward < Tuple > ( t ), std::make_index_sequence < sizeof ... ( ParamTypes ) > { } );
}

template < class ... Params >
struct CheckInput;

template < >
struct CheckInput < > {
	static bool checkInput ( const std::shared_ptr < OperationAbstraction > &, unsigned ) {
		throw std::logic_error ( "Out of types to check" );
	}
};

template < class Param, class ... Params >
struct CheckInput < Param, Params ... > {
	static bool checkInput ( const std::shared_ptr < OperationAbstraction > & operation, unsigned index ) {
		if ( index == 0 )
			return ( bool ) std::dynamic_pointer_cast < typename std::decay < Param >::type > ( operation->getProxyAbstraction ( ) );
		else
			return CheckInput < Params ... >::checkInput ( operation, index - 1 );
	}
};

template < class ... Params >
struct ParamType;

template < >
struct ParamType < > {
	static ext::type_index paramType ( unsigned ) {
		throw std::logic_error ( "Out of types to check" );
	}
};

template < class Param, class ... Params >
struct ParamType < Param, Params ... > {
	static ext::type_index paramType ( unsigned index ) {
		if ( index == 0 )
			return ext::type_index ( typeid ( typename std::decay < Param >::type ) );
		else
			return ParamType < Params ... >::paramType ( index - 1);
	}
};

} /* namespace abstraction */

#endif // _ABSTRACTION_HELPERS_HPP_
